"""Asyncify ChemCloud
Asyncified components:
    - Result request
    - Result polling sleep
    - Compute request
TODO:
    - Refresh token?
"""
from chemcloud.models import (
    FutureResultBase,
    FutureResult,
    FutureResultGroup,
    AtomicInputOrList,
    PossibleResultsOrList,
    GROUP_ID_PREFIX,
)
from chemcloud.http_client import _RequestsClient
from chemcloud.client import CCClient
from chemcloud.utils import _b64_to_bytes, _bytes_to_b64

import trio
import httpx
from qcelemental.util.serialization import json_dumps
from time import time
from typing import Any


async def aget(
    self: FutureResult | FutureResultGroup,
    timeout: float | None = None,
    interval: float = 1.0,
) -> PossibleResultsOrList:
    if self.result:
        return self.result

    start_time = time()
    fid = self.id
    if type(self) is FutureResultGroup:
        fid = fid.replace(GROUP_ID_PREFIX, "")

    while True:
        _, self.result = await self.client.aresult(fid)
        if self.result:
            break
        if timeout:
            if (time() - start_time) > timeout:
                raise TimeoutError(
                    f"Your timeout limit of {timeout} seconds was exceeded"
                )
        await trio.sleep(interval)

    return self.result


FutureResultBase.aget = aget


async def _arequest(
    self,
    method: str,
    route: str,
    *,
    headers: dict[str, str] | None = None,
    data: dict[str, Any] | None = None,
    params: dict[str, Any] | None = None,
    api_call: bool = True,
):
    """Make HTTP request"""
    url = (
        f"{self._chemcloud_domain}"
        f"{self._settings.chemcloud_api_version_prefix if api_call else ''}{route}"
    )
    request = httpx.Request(
        method,
        url,
        headers=headers,
        data=data,
        params=params,
    )
    async with httpx.AsyncClient(timeout=httpx.Timeout(5.0, read=20.0)) as client:
        response = await client.send(request)
    response.raise_for_status()
    return response.json()


async def _authenticated_arequest(self, method: str, route: str, **kwargs):
    """Make authenticated HTTP request"""
    kwargs["headers"] = kwargs.get("headers", {})
    access_token = self._get_access_token()
    kwargs["headers"]["Authorization"] = f"Bearer {access_token}"
    return await self._arequest(
        method,
        route,
        **kwargs,
    )


async def acompute(
    self, input_data: AtomicInputOrList, engine: str, queue: str | None = None
) -> FutureResult | FutureResultGroup:
    """Submit a computation to ChemCloud"""
    # Convery any bytes data to b64 encoding
    _bytes_to_b64(input_data)

    result_id = await self._authenticated_arequest(
        "post",
        "/compute",
        data=json_dumps(input_data),
        params={"engine": engine, "queue": queue},
    )
    return self._result_id_to_future_result(input_data, result_id)


async def aresult(
    self,
    result_id: str,
) -> tuple[str, PossibleResultsOrList | None]:
    result = await self._authenticated_arequest("get", f"/compute/result/{result_id}")
    if result["result"]:
        # Convery b64 encoded native_files to bytes
        _b64_to_bytes(result["result"])

    return result["state"], result["result"]


_RequestsClient._arequest = _arequest
_RequestsClient._authenticated_arequest = _authenticated_arequest
_RequestsClient.acompute = acompute
_RequestsClient.aresult = aresult


async def acompute(
    self, input_data: AtomicInputOrList, engine: str, queue: str | None = None
) -> FutureResult | FutureResultGroup:
    if self.supported_engines is not None:
        assert (
            engine in self.supported_engines
        ), f"Please use one of the following engines: {self.supported_engines}"

    return await self._client.acompute(input_data, engine, queue)


CCClient.acompute = acompute



from tcpb_trio import TCPBClient
import trio

from contextlib import asynccontextmanager
from tempfile import TemporaryDirectory
from random import randrange
from functools import partial
from os import environ


GRAIN_SWORKER_CONFIG = dict()


@asynccontextmanager
async def grain_context(host, port):
    async with TCPBClient(host, port) as tc:
        yield tc


@asynccontextmanager
async def grain_run_sworker():
    with TemporaryDirectory(prefix="tcserver") as tempd:
        async with trio.open_nursery() as _n:
            host, port = trio.socket.gethostname(), randrange(10000, 65536)
            tcs = await _n.start(
                partial(trio.run_process, ["terachem", "-s", str(port)], cwd=tempd)
            )
            print("Waiting for TeraChem server to start...")
            await service_check(host, port)
            yield dict(
                name=f"TCPB@{host}:g={environ.get('CUDA_VISIBLE_DEVICES', 'all')}",
                host=host,
                port=port,
            )
            tcs.terminate()


async def service_check(host, port):
    while True:
        await trio.sleep(1)
        try:
            async with TCPBClient(host, port) as tc:
                while True:
                    if await tc.is_available():
                        return
                    await trio.sleep(1)
        except:
            pass



import subprocess
from os import environ
import argparse
import math


def tcpb_spawn():
    argp = argparse.ArgumentParser(
        epilog="Start TCPB Grain specialized worker based on current machine's GPU resources"
    )
    argp.add_argument(
        "-n",
        "--ngpu",
        type=int,
        default=math.inf,
        help="Maximum number of GPUs to take (default: all of the available ones)",
    )
    args = argp.parse_args()
    _tcpb_spawn(args)


def _tcpb_spawn(args):
    def _run(cmd):
        return subprocess.run(
            cmd, shell=True, check=True, stdout=subprocess.PIPE
        ).stdout.decode()

    total = int(_run("nvidia-smi -L | wc -l").strip())
    nv_smi = iter(_run("nvidia-smi").split("\n"))
    for l in nv_smi:
        if "Processes" in l:
            break
    next(nv_smi)
    next(nv_smi)
    next(nv_smi)
    occupied = set()
    for l in nv_smi:
        if "---" in l or "No" in l:
            break
        occupied.add(int(l.split()[1]))

    l = list(set(range(total)) - occupied)
    l = l[: min(len(l), args.ngpu)]
    print(
        f"{total-len(occupied)} out of {total} GPU(s) are available; taking {len(l)} GPU(s)"
    )
    for i in l:
        p = subprocess.Popen(
            ["grain", "up"],
            env=environ | dict(CUDA_VISIBLE_DEVICES=str(i)),
            start_new_session=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )



import setuptools

exec(open("tc_grain/_version.py").read())

setuptools.setup(
    name="tc-grain",
    version=__version__,
    packages=setuptools.find_packages(),
    python_requires=">=3.10",
    install_requires=[
        "chemcloud",
        "tcpb-trio",
        "trio",
    ],
    extras_require={
        "dev": [
            "black",
            "pre-commit",
        ],
    },
    tests_require=[],
    entry_points={
        "console_scripts": ["tcpb_grain_spawn=tc_grain.util:tcpb_spawn"],
    },
)



from . import chemcloud_trio_patch
from chemcloud import CCClient

from contextlib import asynccontextmanager


GRAIN_SWORKER_CONFIG = dict(
    BACKENDLESS=True,
)


@asynccontextmanager
async def grain_context():
    # Give the tasklets access to the client singleton
    ccclient = CCClient()
    yield ccclient


@asynccontextmanager
async def grain_run_sworker():
    # We don't actually run anything on the ground
    yield dict(
        name="ChemCloud",
    )
